# chrono - agendas system
# Copyright (C) 2017  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import bisect


class Interval(object):
    __slots__ = ['begin', 'end', 'data']

    def __init__(self, begin, end, data=None):
        assert begin < end
        self.begin = begin
        self.end = end
        self.data = data

    def overlap(self, begin, end):
        if end <= self.begin:
            return False
        if begin >= self.end:
            return False
        return True

    def __repr__(self):
        return '<Interval [%s, %s] %s>' % (self.begin, self.end, self.data or '')


class Intervals(object):
    "Maintain a list of mostly non overlapping intervals, allow removing overlap"
    def __init__(self):
        self.points = []
        self.container = []

    def __insert_point(self, point, interval):
        i = bisect.bisect_left(self.points, point)
        if i >= len(self.container) or self.points[i] != point:
            self.points.insert(i, point)
            self.container.insert(i, [])
        self.container[i].append(interval)

    def add(self, begin, end, data=None):
        'Add an interval'
        self.add_interval(Interval(begin, end, data))

    def add_interval(self, interval):
        'Add an interval object'
        self.__insert_point(interval.begin, interval)
        self.__insert_point(interval.end, interval)

    def __iter_interval(self, begin, end, modify=False):
        i = bisect.bisect_left(self.points, begin)
        while i < len(self.points) and self.points[i] <= end:
            container = self.container[i]
            if modify:
                container = list(container)
            for interval in container:
                yield self.points[i], interval
            i += 1

    def remove_overlap(self, begin, end):
        'Remove all overlapping intervals'
        for point, interval in self.__iter_interval(begin, end, modify=True):
            if interval.overlap(begin, end):
                self.__remove_interval(interval)

    def overlap(self, begin, end):
        'Test if some intervals overlap'
        for point, interval in self.__iter_interval(begin, end):
            if interval.overlap(begin, end):
                return True
        return False

    def search(self, begin, end):
        'Search overlapping intervals'
        for point, interval in self.__iter_interval(begin, end):
            if interval.overlap(begin, end):
                # prevent returning the same interval twice
                if interval.begin < begin or interval.begin == point:
                    yield interval

    def search_data(self, begin, end):
        'Search data elements of overlapping intervals'
        for interval in self.search(begin, end):
            yield interval.data

    def iter(self):
        'Iterate intervals'
        if not self.points:
            return []
        return self.search(self.points[0], self.points[-1])

    def iter_data(self):
        'Iterate data element attached to intervals'
        for interval in self.iter():
            yield interval.data

    def __remove_interval(self, interval):
        self.__remove_point_interval(interval.begin, interval)
        self.__remove_point_interval(interval.end, interval)

    def __remove_point_interval(self, point, interval):
        i = bisect.bisect_left(self.points, point)
        assert self.points[i] == point
        self.container[i].remove(interval)
